/**
 * Making DataTables fun again.
 *
 * @param {object} $table
 * @param {object} userOptions
 * @param {object} eventOptions
 * @param {object} translations
 *
 * @return {object}
 */
window.DataTable = function($table, userOptions, eventOptions, translations) {
    'use strict';

    var version = '0.0.9';

    var elements = {
        columnRowSelector:  '.js-table-columns',
        filterRowSelector:  '.js-table-filters',
        filterSelectColumn: '.js-select-filter'
    };

    var prefix = {
        throw: 'w2wDataTables: '
    };

    var defaultOptions = {
        language: 'en',
        stateSaving: true
    };

    var globals = {
        options: $.extend({}, defaultOptions, userOptions || {}),
        source: $table.data('source'),
        autoReload: $table.data('auto-reload'),
        perPage: $table.data('per-page'),
        tableID: $table.attr('id'),
        serverSide: true,
        table: false,
        state: false
    };


    var functions = {
        /**
         * Check if all fields are ok.
         * Create the datatable.
         */
        init: function() {
            functions.hasRequirementsOrThrow();
            functions.makeTable();
            functions.getEventOptions();
        },

        /**
         * Get the custom event options, and merge with the default the events.
         */
        getEventOptions: function() {
            $.each(eventOptions, function(eventKey, eventOption) {
                events[eventKey] = $.extend({}, events[eventKey], eventOption || {});
            });
        },

        /**
         * Check if all fields are ok.
         */
        hasRequirementsOrThrow: function() {
            var columnRow = $table.find(elements.columnRowSelector);

            if (typeof globals.source === 'undefined' || globals.source === '') {
                globals.serverSide = false;
            }

            if (typeof globals.tableID === 'undefined' || globals.tableID === '') {
                throw prefix.throw + 'missing id attribute!';
            }

            if (columnRow.length === 0) {
                throw prefix.throw + 'missing column row (' + elements.columnRowSelector + ')!';
            }

            if (typeof globals.autoReload !== 'undefined' && !globals.autoReload > 0) {
                throw prefix.throw + 'invalid reload interval!';
            }
            if (typeof globals.perPage !== 'undefined' && !globals.perPage > 0) {
                throw prefix.throw + 'invalid amount per page!';
            }
        },

        /**
         * Create the datatable.
         */
        makeTable: function() {
            var tableColumns = functions.getColumns();
            var tableOrder = functions.getOrder();
            var tableLanguage = translations.get(globals.options.language);

            var dataTableConfig = {
                autoWidth: false,
                columns: tableColumns,
                initComplete: function() {
                    var filterRow = $table.find(elements.filterRowSelector);

                    if (filterRow.length > 0) {
                        functions.filterColumn();
                    }
                },
                language: tableLanguage,
                order: tableOrder,
                orderCellsTop: true,
                processing: true,
                responsive: true,
                serverSide: globals.serverSide,
                stateSave: globals.options.stateSaving
            };

            if(globals.serverSide == true) {
                dataTableConfig.ajax = {
                    method: 'POST',
                    url: globals.source
                };
            }

            // eslint-disable-next-line new-cap
            globals.table = $table.DataTable(dataTableConfig);

            globals.state = globals.table.state.loaded();

            functions.setFilterValues();

            if (typeof globals.autoReload !== 'undefined') {
                functions.bindReload(globals.autoReload);
            }

            if (typeof globals.perPage !== 'undefined') {
                functions.setPageLength();
            }

            globals.table.on('init.dt', function() {
                functions.triggerEvent('init.dt');
            });

            // once the table has been drawn, ensure a responsive reculcation
            // if we do not do this, pagination might cause columns to go outside the table
            $.each(events, functions.listenToEvent);
        },

        /**
         * Add a new event.
         *
         * @param {string} on
         * @param {Function} fn
         */
        addEvent: function(on, fn) {
            if(!events[on]) {
                events[on] = [];
            }

            events[on].push(fn);
        },

        /**
         * Listen to an event.
         *
         * @param {string} eventKey
         */
        listenToEvent: function(eventKey) {
            globals.table.on(eventKey, function() {
                functions.triggerEvent(eventKey);
            });
        },

        /**
         * Trigger an event.
         *
         * @param {key} on
         */
        triggerEvent: function(on) {
            if(!events[on]) {
                return;
            }

            $.each(events[on], function(index, fn) {
                fn();
            });
        },

        /**
         * Set the page length.
         */
        setPageLength: function() {
            globals.table.page.len(globals.perPage).draw();
        },

        /**
         * Get the columns.
         *
         * @return {array}
         */
        getColumns: function() {
            var tableColumns = $table.find(elements.columnRowSelector + ' th');
            var columns = [];

            tableColumns.each(function() {
                // set default options
                var defOrderable = true;
                var defSearchable = true;
                var validOptionsSortOrder = [
                    true,
                    false
                ];
                // get the column values
                var column = $(this);
                var columnName = column.data('name');
                var columnOrderable = column.data('orderable');
                var columnSearchable = column.data('searchable');

                if (typeof columnOrderable === 'undefined' ||
                    !validOptionsSortOrder.indexOf(columnOrderable)) {
                    columnOrderable = defOrderable;
                }

                if (typeof columnSearchable === 'undefined' ||
                    !validOptionsSortOrder.indexOf(columnSearchable)) {
                    columnSearchable = defSearchable;
                }

                columns.push({
                    data: columnName,
                    name: columnName,
                    orderable: columnOrderable,
                    searchable: columnSearchable
                });
            });

            return columns;
        },

        /**
         * Get the order.
         *
         * @return {array}
         */
        getOrder: function() {
            var defaultOrder = [
                [0, 'desc']
            ];
            var validSortOrders = ['asc', 'desc'];
            var sortColumn = $table.find('[data-default-sort="true"]');
            var sortColumnOrder = sortColumn.data('default-sort-order');

            if (sortColumn.length === 0) {
                // no custom sort column on this table - use the default settings
                return defaultOrder;
            }

            if (typeof sortColumnOrder === 'undefined') {
                throw prefix.throw +
                    'You must add a sorting order (default-sort-order="asc/desc")' +
                    ' if you are filtering on a custom column!';
            }

            if (validSortOrders.indexOf(sortColumnOrder) == -1) {
                throw prefix.throw +
                    'You must add a valid sorting order (asc/desc) if you are filtering on a custom column!';
            }

            return [
                [
                    sortColumn.index(),
                    sortColumnOrder
                ]
            ];
        },

        /**
         * Filter the columns.
         */
        filterColumn: function() {
            globals.table
                .columns()
                .eq(0)
                .each(function(colIdx) {
                    var tableFilter = $table.find(elements.filterRowSelector + ' th:eq(' + colIdx + ')');

                    functions.initFilterSelect(colIdx, tableFilter);
                    functions.initFilterInput(colIdx, tableFilter);
                });
        },

        /**
         * Initialize the input filter.
         *
         * @param {string} colIdx
         * @param {object} tableFilter
         */
        initFilterInput: function(colIdx, tableFilter) {
            var debouncedFiltering = functions.debounce(function(searchValue) {
                globals.table
                    .column(colIdx)
                    .search(searchValue)
                    .draw();
            }, 250);

            tableFilter.find('.js-input-filter').on('input', function() {
                var input = $(this);
                var searchValue = input.val();

                debouncedFiltering(searchValue);
            });
        },

        /**
         * Initialize the select filter.
         *
         * @param {string} colIdx
         * @param {object} tableFilter
         */
        initFilterSelect: function(colIdx, tableFilter) {
            tableFilter.find(elements.filterSelectColumn).on('change', function() {
                var select = $(this);
                var searchValue = select.val();

                globals.table
                    .column(colIdx)
                    .search(searchValue)
                    .draw();
            });
        },

        /**
         * Bind the reload.
         *
         * @param {string} interval
         */
        bindReload: function(interval) {
            setInterval(function() {
                globals.table.ajax.reload();
            }, interval);
        },

        /**
         * Debounce a function.
         *
         * @param {object} func
         * @param {number} wait
         * @param {boolean} immediate
         *
         * @return {object}
         */
        debounce: function(func, wait, immediate) {
            var timeout;

            return function() {
                var context = this;
                var args = arguments;

                clearTimeout(timeout);
                timeout = setTimeout(function() {
                    timeout = null;
                    if (!immediate) {
                        func.apply(context, args);
                    }
                }, wait);
                if (immediate && !timeout) {
                    func.apply(context, args);
                }
            };
        },

        /**
         * Recalc the table.
         */
        recalc: function() {
            globals.table.responsive.recalc();
        },

        /**
         * Set the filter values.
         */
        setFilterValues: function() {
            if(!globals.state.columns) {
                return;
            }

            $.each(globals.state.columns, function(column, value) {
                $(elements.filterRowSelector + ' .form-control').eq(column).val(value.search.search);
            });
        }
    };

    var events = {
        'draw.dt': [
            functions.recalc
        ]
    };

    return {
        options:   globals.options,
        functions: functions,
        element:   $table,
        events:    events,
        version:   version
    };
};
